/*

Copyright (c) 2009, SFB/TR-55
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*
REVISION HISTORY
Date            Name                    Comments
Feb 09 2009     Dieter Hierl            created
Apr 27 2009     Dieter Hierl            version-0.7
Mai 12 2009     Dieter Hierl            version-0.8
May 18 2009     Dieter Hierl            version-1.0
Jun 22 2009     Dieter Hierl            version-1.1
Jul 02 2009     Dieter Hierl            version-1.2
Nov 04 2009     Stefan Solbrig          total rewrite, version 2.0
Jan 12 2010     Stefan Solbrig          total rename, version 3.0
Jan 25 2010     Stefan Solbrig          change counters, version 4.0
Jan 25 2010     Stefan Solbrig          add some static variables, version 4.1
Mar 04 2010     Stefan Solbrig          un-register from async. events, version 4.2.1
*/

#ifndef GS_H
#define GS_H

#ifdef __kernel__
#include <linux/types.h>
#endif

#define DRV_NAME_EXPLICIT "Global Signals Device Driver"
#define DRV_NAME          "qpace-nwp-gs"
#define DRV_VERSION       "v4.2.1"

enum {
	/*
	 * Other flags that can be passed as the second argument to ioctl()
	 */
	GS_WRITE_REGISTER = 0x100, /* for use with ioctl()
	                              and the above registers */
	GS_READ_REGISTER  = 0x200, /* for use with ioctl()
	                              and the above register */
	GS_OTHER          = 0x1000, /* for use with ioctl()
	                               and non-register operations
	                               listed below */
	GS_SET_VERBOSITY  = 0x1001, /* for use with GS_OTHER */
	GS_GET_VERBOSITY  = 0x1002, /*  '' */
	GS_SET_FASYNCERS  = 0x1003, /*  '' */
	GS_GET_FASYNCERS  = 0x1004, /*  '' */
	GS_SET_UNBLOCKERS = 0x1005, /*  '' */
	GS_GET_UNBLOCKERS = 0x1006, /*  '' */
	GS_SET_RELEASEKILL= 0x1007, /*  '' */
	GS_GET_RELEASEKILL= 0x1008, /*  '' */
	GS_BARRIER        = 0x1009,  /*  '' */
	GS_SET_STATIC_OVERFLOW = 0x1010, /* '' */
	GS_SET_STATIC_KILLS    = 0x1011, /* '' */
	GS_SET_STATIC_VALIDS   = 0x1012, /* '' */
	GS_GET_STATIC_OVERFLOW = 0x1013, /* '' */
	GS_GET_STATIC_KILLS    = 0x1014, /* '' */
	GS_GET_STATIC_VALIDS   = 0x1015  /* '' */
};


#define VHDLBIT(x) (0x80000000 >> (x))

#define VHDLBITS(x,y) \
    ( \
      ((0x80000000 >>  0) & ((  0 >= (x) &&  0 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  1) & ((  1 >= (x) &&  1 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  2) & ((  2 >= (x) &&  2 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  3) & ((  3 >= (x) &&  3 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  4) & ((  4 >= (x) &&  4 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  5) & ((  5 >= (x) &&  5 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  6) & ((  6 >= (x) &&  6 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  7) & ((  7 >= (x) &&  7 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  8) & ((  8 >= (x) &&  8 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >>  9) & ((  9 >= (x) &&  9 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 10) & ((  0 >= (x) &&  0 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 11) & (( 11 >= (x) && 11 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 12) & (( 12 >= (x) && 12 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 13) & (( 13 >= (x) && 13 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 14) & (( 14 >= (x) && 14 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 15) & (( 15 >= (x) && 15 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 16) & (( 16 >= (x) && 16 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 17) & (( 17 >= (x) && 17 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 18) & (( 18 >= (x) && 18 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 19) & (( 19 >= (x) && 19 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 20) & (( 20 >= (x) && 20 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 21) & (( 21 >= (x) && 21 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 22) & (( 22 >= (x) && 22 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 23) & (( 23 >= (x) && 23 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 24) & (( 24 >= (x) && 24 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 25) & (( 25 >= (x) && 25 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 26) & (( 26 >= (x) && 26 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 27) & (( 27 >= (x) && 27 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 28) & (( 28 >= (x) && 28 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 29) & (( 29 >= (x) && 29 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 30) & (( 30 >= (x) && 30 <= (y)) ? 0xffffffff : 0x00000000)) |\
      ((0x80000000 >> 31) & (( 31 >= (x) && 31 <= (y)) ? 0xffffffff : 0x00000000))  )

/*
 * GS registers
 */

enum {
	/* 
	 * Status register. Readable. non-ROC, non-COW/
	 */
	GS_STATUS       = 0x0,

	/* same as GS_STATUS, but bits 24, 25, are clear-on-read,
	 * and interrupt gets ack'ed (clears bit 7)
	 */
	GS_STATUS_CLEAR = 0x1
};

enum {
	GS_STATUS_GLOB_EXC   = VHDLBIT(0),  /* Global exception active. */
	GS_STATUS_EXC_WRWR   = VHDLBIT(1),  /* write-write exception active. */
	GS_STATUS_EXC_RXIDLE = VHDLBIT(2),  /* Exception if something is received, 
	                                     * but nothing was sent active. */
	GS_STATUS_EXC_TO     = VHDLBIT(3),  /* Timeout exception active. */
	GS_STATUS_EXC_NOACK  = VHDLBIT(4),  /* Noack exception. */
	/* bits 5 thru 6 are unused */
	GS_STATUS_INTERRUPT  = VHDLBIT(7),  /* Current value of the interrupt line/ */
	/* bits 8 thru 23 are unused */
	GS_STATUS_RX_KILL    = VHDLBIT(24),  /* Kill is being received. */
	GS_STATUS_RX_VALID   = VHDLBIT(25),  /* Valid is being received. */
	GS_STATUS_RX_VALUE   = VHDLBIT(26),  /* Value is being received. */
	/* bit 27 is unused */
	GS_STATUS_TX_KILL    = VHDLBIT(28),  /* Sending kill. */
	GS_STATUS_TX_VALID   = VHDLBIT(29),  /* Sending valid. */
	GS_STATUS_TX_VALUE   = VHDLBIT(30)   /* Sending value. */
	/* bit 31 is unused */
};

enum {
	/*
	 * Command register. Only writabe, not readable.
	 * Writing twice causes the WRWR exception.
	 */
	GS_CMD = 0x2   /* not readable [writable] */
};

enum {
	GS_CMD_TRUE = VHDLBIT(31),  /* writable. write 1 sends true, 
	                            * write 0 sends false */
	GS_CMD_FALSE = 0x0
};

enum {
	/*
	 * Configuration register. Readable and writable. non-COW, non-COR.
	 */
	GS_CONFIG = 0x3 /* r/w, no COW, NO COR */
};

enum {
	GS_CONFIG_GLOB_EXC_EN  = VHDLBIT(0),  /* If global exception occurs, 
	                                       * then send a kill and,
	                                       * notify processor via interrupt
	                                       */
	GS_CONFIG_EXC_WRWR_EN  = VHDLBIT(1),  /* If a wrwr exception occurs, forward it. */
	GS_CONFIG_EXC_IDLE_EN  = VHDLBIT(2),  /* If an idle exception occurs, forward it. */
	GS_CONFIG_EXC_TO_EN    = VHDLBIT(3),  /* If a timeout exception occurs, forward it. */
	GS_CONFIG_EXC_NOACK_EN = VHDLBIT(4),  /* If a noack exception occurs, forward it. */
	GS_CONFIG_INTERRUPT_EN = VHDLBIT(7),  /* Enable interrupt. */
	GS_CONFIG_CLOSED       = VHDLBIT(11), /* Close GS on the card. (Card sends a NOP). */
	GS_CONFIG_TO_LIMIT     = VHDLBITS(18,31),  /* Timeout limit mask. */
	GS_CONFIG_TO_LIMIT_S   = 0x0               /* Timeout limit shift. */
	/* The timeout is approx. z:
	 * z = (2^24 -1)*N 
	 * N is the integer stored in GS_CONFIG_TO_LIMIT
	 */
};

enum {
	/*
	 * Counter Register. Writable and readable.
	 * However, writing to this register does nothing but
	 * clears the interrupt counter.
	 */
	GS_COUNTERS = 0x4
};

enum {
	GS_COUNTERS_TO_COUNTER = VHDLBITS(2,15),   /* 16 most sig. bits of the timeout counter. */
	GS_COUNTERS_TO_COUNTER_S = 0x10,           /* Shift for the timeout counter. */
	GS_COUNTERS_INT_COUNTER = VHDLBITS(16,23), /* Number of interrupts. */
	GS_COUNTERS_INT_COUNTER_S = 0x8,           /* Shift for the interrupt counter. */
	GS_COUNTERS_LAT_COUNTER = VHDLBITS(24,31), /* latency counter */
	GS_COUNTERS_LAT_COUNTER_S = 0x0            /* Shift for the latency counter. */ 
	/* on every write, the interruptcounter gets cleared */
	/*
	 * Timeout counter runs if valid or kill has been sent.
	 * Gets reset if something is received.
	 * Counter runs if something has been send and nothing is being received.
	 *
	 * Latency counter is 12 bits internally, one reads the 8 MSB.
	 */
};

/*
 * Use this macro like this:
 * GS_COMPUTE_COUNTER(INT, value);
 * GS_COMPUTE_COUNTER(TO, value);
 * GS_COMPUTE_COUNTER(LAT, value);
 * with "value" being a 32-bit value read
 * from the GS_COUNTERS register.
 */
#define GS_COMPUTE_COUNTER(kind,value) \
    (((value) & GS_COUNTERS_ ##kind ## _COUNTER) >> GS_COUNTERS_ ##kind ## _COUNTER_S)


enum {
	/*
	 * Global Exceptions register. Only readable.
	 */
	GS_GLOB_EXCS = 0x5
};

enum {
	GS_GLOB_EXCS_EXC_OPCCC    = VHDLBIT(0),
	GS_GLOB_EXCS_EXC_IWC      = VHDLBIT(1),
	GS_GLOB_EXCS_EXC_OWC      = VHDLBIT(2),
	GS_GLOB_EXCS_EXC_ETH      = VHDLBIT(3),
	/* bits 4 and 5 are unused */
	GS_GLOB_EXCS_EXC_TORUS_TX5 = VHDLBIT(6),
	GS_GLOB_EXCS_EXC_TORUS_TX4 = VHDLBIT(7),
	GS_GLOB_EXCS_EXC_TORUS_TX3 = VHDLBIT(8),
	GS_GLOB_EXCS_EXC_TORUS_TX2 = VHDLBIT(9),
	GS_GLOB_EXCS_EXC_TORUS_TX1 = VHDLBIT(10),
	GS_GLOB_EXCS_EXC_TORUS_TX0 = VHDLBIT(11),
	/* bits 12 and 13 are unused */
	GS_GLOB_EXCS_EXC_TORUS_RX5 = VHDLBIT(14),
	GS_GLOB_EXCS_EXC_TORUS_RX4 = VHDLBIT(15),
	GS_GLOB_EXCS_EXC_TORUS_RX3 = VHDLBIT(16),
	GS_GLOB_EXCS_EXC_TORUS_RX2 = VHDLBIT(17),
	GS_GLOB_EXCS_EXC_TORUS_RX1 = VHDLBIT(18),
	GS_GLOB_EXCS_EXC_TORUS_RX0 = VHDLBIT(19),
	GS_GLOB_EXCS_EXC_GS        = VHDLBIT(20),
	GS_GLOB_EXCS_EXC_SOFT      = VHDLBIT(21)
};

enum {
	/*
	 * GS version regiser. Only readable.
	 */
	GS_VERSION = 0x6
};

enum {
	GS_VERSION_MAJOR_MASK = 0xfff00000,
	GS_VERSION_MINOR_MASK = 0x000f0000,
	GS_VERSION_MICRO_MASK = 0x0000ffff,

	GS_VERSION_MAJOR_SHIFT = 20,
	GS_VERSION_MINOR_SHIFT = 16,
	GS_VERSION_MICRO_SHIFT = 0
};

/*
 * Use this macro like this:
 * GS_COMPUTE_VERSION(MAJOR, value)
 * GS_COMPUTE_VERSION(MINOR, value)
 * GS_COMPUTE_VERSION(MICRO, value)
 * with "value" being a 32-bit value read
 * from the GS_VERSION register
 */
#define GS_COMPUTE_VERSION(kind,value) \
    (((value) & GS_VERSION_ ##kind ## _MASK) >> GS_VERSION_ ##kind ## _SHIFT)

enum {
	/*
	 * Reset register. Writable.
	 */
	GS_RESET     = 0x11
};

enum {
	GS_RESET_DOWN = 0x1,  /* Puts the GS into reset. */
	GS_RESET_LIFT = 0x0   /* Get the GS out of reset. */
};

enum {
	GS_SOFTEXC   = 0x16,   /* trigger soft exception */
};

enum {
	GS_SOFTEXC_ACTIVE = 0x00,
	GS_SOFTEXC_CLEAR =  0x01
};

/*
 * Device numbers
 */
#define GS_DEVSIZE      0
#define GS_MAJOR      211
#define GS_MINOR        0

#define GS_DEVICE_COUNT 1

enum {
	GS_TRUE =  1,
	GS_FALSE = 0
};

#ifdef __KERNEL__

/*
 * Internal data structure used by the device driver.
 * Not to be used outside of the kernel.
 */
struct gs_device {
	/* use to lock access to shared static variable,
	 * use e.g., to guard access to gs_last_interrupt
	 * gs_int_overflow and gs_last_interrupt.
	 */
	spinlock_t    gs_spin_lock;

	/* use to lock access to static variables in
	 * sections that can sleep, e.g., access to
	 * gs_last_signal or gs_have_signal in read()
	 */
	struct semaphore gs_mutex;

	/*
	 * Both fields contain bits that are checked
	 * against bits from the GS_INTERRUPT register.
	 * If the bit is set in .gs_fasyncers and GS_INTERRUPT,
	 * an interrupts wakes up all asynchronous readers
	 * that own the file descriptor.
	 * If the bit is set in .gs_unblockers, all processes
	 * that wait on a blocking read() are unblocked.
	 */
	u32           gs_fasyncers;
	u32           gs_unblockers;

	/* wait queue to manage blocking read()/write () */
	wait_queue_head_t gs_waitqueue;

	/* signal overflow counter, incremented if the
	 * tasklet cannot handle incoming signals fast enough.
	 */
	u64           gs_signal_overflow;

	/* user reads this via read() */
	unsigned char gs_last_signal;

	/* set to 1 if  gs_last_signal is in a sensible state
	 * and has not yet been consumed by the user
	 */
	unsigned int  gs_have_signal;

	/* pointer to list of asynchronous readers */
	struct fasync_struct *gs_fasyncerp;

	/* dcr map returned by firmware helper functions */
	dcr_host_t    gs_dcr_map;

	/* interrupt number */
	int           gs_int_num;

	/*our character device */
	struct cdev   gs_chardev;

	/* number of clients */
	unsigned int gs_clients;

	/* number of valids */
	u64 gs_valids;

	/* number of kills */
	u64 gs_kills;
};

struct gs_private {
	struct gs_device *gsp_devp;
	int gsp_releasekill;
};

#endif /* defined __KERNEL__ */

#endif /* defined GS_H */


